// [[Rcpp::depends(RcppArmadillo)]]
#include <iostream> 
#include <RcppArmadillo.h>
#include <stdlib.h> 
#include "util.h"






// [[Rcpp::export]]
Rcpp::List dfa_without_prior_binary(arma::mat z, int N, int burnin, int thin){//main function
  arma::uword i,j,k,it,jj,kk,nb,np,nk;//iterators
  arma::uvec iv(1);
  
  int N2 = N / 10;
  int burnin2 = N2 / 2;
  int thin2 = thin;
  int Kmax=50;
  
  double a_beta=1;
  double b_beta=1;
  double w_sig=10;
  double alpha=1;
  double w_sd=1.5;
  double zeta_sd=.8;
  double zeta_sig=100;

  int n = z.n_rows;
  int p = z.n_cols;
  int effN = (N-burnin)/thin;
  arma::vec K(effN,arma::fill::zeros);
  arma::cube P = arma::zeros(n,Kmax,effN);
  
  int K_old=1;
  arma::mat P_old=arma::zeros(n,Kmax);
  
  arma::mat B_old=arma::zeros(p,Kmax);
  arma::mat w_B_old(p,Kmax,arma::fill::ones);
  arma::vec zeta_old(p,arma::fill::zeros);
  zeta_old.fill(-5);
  
  double beta_old=.2;
  arma::mat PB_kernel_old=arma::zeros(n,p);
  arma::mat PB_kernel_new=arma::zeros(n,p);
  
  arma::mat ll_z_old=arma::zeros(n,p);
  arma::mat ll_z_new=ll_z_old;
  
  for (i=0;i<n;i++){
    for (j=0;j<p;j++){
      PB_kernel_old(i,j)=sum(w_B_old.row(j)%P_old.row(i)%B_old.row(j))+zeta_old(j);
      ll_z_old(i,j)=-logsumexp0(PB_kernel_old(i,j))+z(i,j)*PB_kernel_old(i,j);
    }
  }
  double ll0=accu(ll_z_old);
  arma::vec ll(N,arma::fill::zeros);
  int iter=0;
  double ac_w=0,ac_zeta=0;
  int count_w=0,count_zeta=0;
  
  arma::rowvec sum_P(Kmax);
  double log1,log0,logp,lr,w_new,zeta_new,d;
  arma::uvec actB,actP,actInd,inactP,inactInd;
  int K_new;
  arma::vec w_B_new(p);
  arma::vec B_new(p);
  //MCMC starts here
  
  for (it=0;it<N;it++){
    Rcpp::checkUserInterrupt();

    //update P
    PB_kernel_new=PB_kernel_old;
    ll_z_new=ll_z_old;

    for (i=0;i<n;i++){
      iv(0)=i;
      //update existing P features with latent disease
      if (K_old>0){
        sum_P = sum(P_old,0)-P_old.row(i);
        actP=find(sum_P>0); //active diseases
        np=actP.n_elem;
        inactP=find(sum_P==0); //inactive diseases
        P_old(iv,inactP).fill(0); //wipe out all inactive diseases
        for (kk=0;kk<np; kk++){
          k=actP(kk);
          log1= log(sum_P(k));
          log0= log(n-sum_P(k));
          actB = find(B_old.col(k)!=0); //active binary symtoms for disease k
          nb=actB.n_elem;

          for (jj=0;jj<nb;jj++){
            j=actB(jj);
            if (P_old(i,k)==1){
              PB_kernel_new(i,j)=PB_kernel_old(i,j)-w_B_old(j,k);
              ll_z_new(i,j)=-logsumexp0(PB_kernel_new(i,j));
              if (z(i,j)==1){
                ll_z_new(i,j)+=PB_kernel_new(i,j);
              }
              log1+=ll_z_old(i,j);
              log0+=ll_z_new(i,j);
            }else{
              PB_kernel_new(i,j)=PB_kernel_old(i,j)+w_B_old(j,k);
              ll_z_new(i,j)=-logsumexp0(PB_kernel_new(i,j));
              if (z(i,j)==1){
                ll_z_new(i,j)+=PB_kernel_new(i,j);
              }
              log0+=ll_z_old(i,j);
              log1+=ll_z_new(i,j);
            }
          }
          

          logp = log1 - logsumexp(log1, log0);
          if (log(R::runif(0,1)) < logp) {
            if (P_old(i,k)==0){
              P_old(i,k) = 1;
              PB_kernel_old.row(i)=PB_kernel_new.row(i);
              ll_z_old.row(i)=ll_z_new.row(i);
            }else{
              PB_kernel_new.row(i)=PB_kernel_old.row(i);
              ll_z_new.row(i)=ll_z_old.row(i);
            }
          } else {
            if (P_old(i,k)==1){
              P_old(i,k) = 0;
              PB_kernel_old.row(i)=PB_kernel_new.row(i);
              ll_z_old.row(i)=ll_z_new.row(i);
            }else{
              PB_kernel_new.row(i)=PB_kernel_old.row(i);
              ll_z_new.row(i)=ll_z_old.row(i);
            }
          }
        }
      }
      
      //update new features with latent disease
      inactInd=find(sum(P_old,0)==0);
      K_old=Kmax-inactInd.n_elem;
      if (K_old<Kmax){
        K_new=R::rbinom(1,alpha/n);//min(rpois(1,alpha/n),Kmax-K_old)
        if (K_new==1){
          B_new=Rcpp::rbinom(p,1,beta_old);
          actB = find(B_new!=0);
          nb=actB.n_elem;
          w_B_new.fill(0);
          
          if (nb>0){
            lr=0;
            for (jj=0;jj<nb;jj++){
              j=actB(jj);
              w_B_new(j)=R::rgamma(1.0,w_sig);
              PB_kernel_new(i,j)=PB_kernel_old(i,j)+w_B_new(j);
              ll_z_new(i,j)=-logsumexp0(PB_kernel_new(i,j))+z(i,j)*PB_kernel_new(i,j);
              lr += ll_z_new(i,j)-ll_z_old(i,j);
            }
            
            lr += R::dpois(K_new,alpha/n,1)-R::dpois(0,alpha/n,1)+R::dbinom(0,1,alpha/n,1)-R::dbinom(K_new,1,alpha/n,1);
            
            if (lr > log(R::runif(0,1))) {
              kk=inactInd(0);
              K_old++;
              w_B_old.col(kk)=w_B_new;
              B_old.col(kk)=B_new;
              PB_kernel_old.row(i)=PB_kernel_new.row(i);
              P_old(i,kk)=1;
              ll_z_old.row(i)=ll_z_new.row(i);
            }else{
              PB_kernel_new.row(i)=PB_kernel_old.row(i);
              ll_z_new.row(i)=ll_z_old.row(i);
            }
          }
        }
      }
    }
    Rcpp::checkUserInterrupt();
    
    actInd=find(sum(P_old,0)>0);
    nk=actInd.n_elem;
    //update B
    for (j=0;j<p;j++){
      for (kk=0;kk<nk;kk++){
        k=actInd(kk);
        log1=log(beta_old);
        log0=log(1-beta_old);
        if (B_old(j,k)==1){//propose to turn off
          for (i=0;i<n;i++){
            log1+=ll_z_old(i,j);
            PB_kernel_new(i,j)=PB_kernel_old(i,j)-w_B_old(j,k)*P_old(i,k);
            ll_z_new(i,j)=z(i,j)*PB_kernel_new(i,j)-logsumexp0(PB_kernel_new(i,j));
            log0+=ll_z_new(i,j);
          }
        }else{//propose to turn on
          w_new=R::rgamma(1.0,w_sig);
          for (i=0;i<n;i++){
            PB_kernel_new(i,j)=PB_kernel_old(i,j)+w_new*P_old(i,k);
            ll_z_new(i,j)=z(i,j)*PB_kernel_new(i,j)-logsumexp0(PB_kernel_new(i,j));
            log0+=ll_z_old(i,j);
            log1+=ll_z_new(i,j);
          }
        }
        //draw binomial
        logp = log1 - logsumexp(log1, log0);
        if (log(R::runif(0,1)) < logp){
          if (B_old(j,k)==0){
            B_old(j,k)=1;
            PB_kernel_old.col(j)=PB_kernel_new.col(j);
            ll_z_old.col(j)=ll_z_new.col(j);
            w_B_old(j,k)=w_new;
          }
        }else{
          if (B_old(j,k)==1){
            B_old(j,k)=0;
            PB_kernel_old.col(j)=PB_kernel_new.col(j);
            ll_z_old.col(j)=ll_z_new.col(j);
            w_B_old(j,k)=0;
          }
        }
      }
    }
    
    //update w_B;
    for (j=0;j<p;j++){
      for (kk=0;kk<nk;kk++){
        k=actInd(kk);
        if (B_old(j,k)==1){
          count_w++;
          w_new=R::rlnorm(log(pow(w_B_old(j,k),2)/sqrt(pow(w_sd,2)+pow(w_B_old(j,k),2))),sqrt(log(pow(w_sd,2)/pow(w_B_old(j,k),2)+1)));
          for (i=0;i<n;i++){
            PB_kernel_new(i,j)=PB_kernel_old(i,j)+(w_new-w_B_old(j,k))*P_old(i,k);
            ll_z_new(i,j)=z(i,j)*PB_kernel_new(i,j)-logsumexp0(PB_kernel_new(i,j));
          }
          lr = sum(ll_z_new.col(j))-sum(ll_z_old.col(j))+R::dgamma(w_new,1,w_sig,1)-R::dgamma(w_B_old(j,k),1,w_sig,1)+R::dlnorm(w_B_old(j,k),log(pow(w_new,2)/sqrt(pow(w_sd,2)+pow(w_new,2))),sqrt(log(pow(w_sd,2)/pow(w_new,2)+1)),1)-R::dlnorm(w_new,log(pow(w_B_old(j,k),2)/sqrt(pow(w_sd,2)+pow(w_B_old(j,k),2))),sqrt(log(pow(w_sd,2)/pow(w_B_old(j,k),2)+1)),1);
          if (lr>log(R::runif(0,1))){
            ac_w++;
            w_B_old(j,k)=w_new;
            PB_kernel_old.col(j)=PB_kernel_new.col(j);
            ll_z_old.col(j)=ll_z_new.col(j);
          }
        }
      }
    }
    
    
    //update zeta
    for (j=0;j<p;j++){
      count_zeta++;
      zeta_new=R::rnorm(zeta_old(j),zeta_sd);
      for (i=0;i<n;i++){
        PB_kernel_new(i,j)=PB_kernel_old(i,j)+zeta_new-zeta_old(j);
        ll_z_new(i,j)=z(i,j)*PB_kernel_new(i,j)-logsumexp0(PB_kernel_new(i,j));
      }
      lr = sum(ll_z_new.col(j))-sum(ll_z_old.col(j))+R::dnorm4(zeta_new,log(1.0/99),zeta_sig,1)-R::dnorm4(zeta_old(j),log(1.0/99),zeta_sig,1);
      if (lr>log(R::runif(0,1))){
        ac_zeta++;
        zeta_old(j)=zeta_new;
        PB_kernel_old.col(j)=PB_kernel_new.col(j);
        ll_z_old.col(j)=ll_z_new.col(j);
      }
    }
    
    //remove empty features
    inactInd=find(my_NAND(sum(P_old,0),sum(B_old,0)));//double check
    P_old.cols(inactInd).fill(0);
    B_old.cols(inactInd).fill(0);
    w_B_old.cols(inactInd).fill(0);
    K_old=Kmax-inactInd.n_elem;
    
    
    //update beta
    d=accu(B_old);
    beta_old=R::rbeta(a_beta+d,b_beta+p*K_old-d);
    
    ll(it)=accu(ll_z_old);
    if (it%1000==0){
      Rcpp::Rcout<<"MCMC: "<<it<<"/"<<N<<" completed"<<std::endl;
    }
    //write
    if (it>=burnin&&it%thin==0){
      P.slice(iter)=P_old;
      K(iter)=K_old;
      iter++;
    }
  }//end of MCMC
  ac_w/=count_w;
  ac_zeta/=count_zeta;
  arma::vec K_freq=K_FREQ(K);
  arma::uword K_est=index_max(K_freq);
  if (K_est==0){
    Rcpp::Rcout<<"Program stops because K_est=0. No resutls returned."<<std::endl;
    return Rcpp::List::create(Rcpp::Named("K")=K,Rcpp::Named("K_est")=K_est,Rcpp::Named("ll")=ll,Rcpp::Named("ll0")=ll0,Rcpp::Named("P")=P,Rcpp::Named("ac_w")=ac_w,Rcpp::Named("ac_zeta")=ac_zeta);
  }
  arma::uvec K_ind=find(K==K_est);
  
  arma::cube P_tmp(n,Kmax,K_ind.n_elem);
  for (i=0;i<K_ind.n_elem;i++){
    P_tmp.slice(i)=P.slice(K_ind(i));
  }
  double K_choose = MAP(P_tmp);
  
  arma::mat P_est = P_tmp.slice(K_choose);
  
  P_est = P_est.cols(sort_index(sum(P_est,0),"descend"));
  
  P_est= P_est.cols(0,K_est-1);
  //P_est=P_est.cols(find(sum(P_est,0)>=.05*n));//Uncomment this line to keep diseases with at least 5% patients

  
  
  
  int effN2 = (N2-burnin2)/thin2;
  arma::cube B=arma::zeros(p,K_est,effN2);
  arma::vec beta(effN2,arma::fill::zeros);
  arma::cube w_B=arma::zeros(p,K_est,effN2);
  arma::mat zeta=arma::zeros(p,effN2);

  
  
  B_old=arma::zeros(p,K_est);
  w_B_old=arma::ones(p,K_est);
  for (i=0;i<n;i++){
    for (j=0;j<p;j++){
      PB_kernel_old(i,j)=sum(w_B_old.row(j)%P_est.row(i)%B_old.row(j))+zeta_old(j);
      ll_z_old(i,j)=-logsumexp0(PB_kernel_old(i,j))+z(i,j)*PB_kernel_old(i,j);
    }
  }
  arma::vec ll2(N2,arma::fill::zeros);
  iter=0;
  double ac_w2=0,ac_zeta2=0;
  int count_w2=0,count_zeta2=0;
  
  for (it=0;it<N2;it++){
    Rcpp::checkUserInterrupt();
    
    PB_kernel_new=PB_kernel_old;
    ll_z_new=ll_z_old;
    //update B
    for (j=0;j<p;j++){
      for (k=0;k<K_est;k++){
        log1=log(beta_old);
        log0=log(1-beta_old);
        if (B_old(j,k)==1){//propose to turn off
          for (i=0;i<n;i++){
            log1+=ll_z_old(i,j);
            PB_kernel_new(i,j)=PB_kernel_old(i,j)-w_B_old(j,k)*P_est(i,k);
            ll_z_new(i,j)=z(i,j)*PB_kernel_new(i,j)-logsumexp0(PB_kernel_new(i,j));
            log0+=ll_z_new(i,j);
          }
        }else{//propose to turn on
          w_new=R::rgamma(1.0,w_sig);
          for (i=0;i<n;i++){
            PB_kernel_new(i,j)=PB_kernel_old(i,j)+w_new*P_est(i,k);
            ll_z_new(i,j)=z(i,j)*PB_kernel_new(i,j)-logsumexp0(PB_kernel_new(i,j));
            log0+=ll_z_old(i,j);
            log1+=ll_z_new(i,j);
          }
        }
        //draw binomial
        logp = log1 - logsumexp(log1, log0);
        if (log(R::runif(0,1)) < logp){
          if (B_old(j,k)==0){
            B_old(j,k)=1;
            PB_kernel_old.col(j)=PB_kernel_new.col(j);
            ll_z_old.col(j)=ll_z_new.col(j);
            w_B_old(j,k)=w_new;
          }
        }else{
          if (B_old(j,k)==1){
            B_old(j,k)=0;
            PB_kernel_old.col(j)=PB_kernel_new.col(j);
            ll_z_old.col(j)=ll_z_new.col(j);
            w_B_old(j,k)=0;
          }
        }
      }
    }
    
    //update w_B;
    for (j=0;j<p;j++){
      for (k=0;k<K_est;k++){
        if (B_old(j,k)==1){
          count_w2++;
          w_new=R::rlnorm(log(pow(w_B_old(j,k),2)/sqrt(pow(w_sd,2)+pow(w_B_old(j,k),2))),sqrt(log(pow(w_sd,2)/pow(w_B_old(j,k),2)+1)));
          for (i=0;i<n;i++){
            PB_kernel_new(i,j)=PB_kernel_old(i,j)+(w_new-w_B_old(j,k))*P_est(i,k);
            ll_z_new(i,j)=z(i,j)*PB_kernel_new(i,j)-logsumexp0(PB_kernel_new(i,j));
          }
          lr = sum(ll_z_new.col(j))-sum(ll_z_old.col(j))+R::dgamma(w_new,1,w_sig,1)-R::dgamma(w_B_old(j,k),1,w_sig,1)+R::dlnorm(w_B_old(j,k),log(pow(w_new,2)/sqrt(pow(w_sd,2)+pow(w_new,2))),sqrt(log(pow(w_sd,2)/pow(w_new,2)+1)),1)-R::dlnorm(w_new,log(pow(w_B_old(j,k),2)/sqrt(pow(w_sd,2)+pow(w_B_old(j,k),2))),sqrt(log(pow(w_sd,2)/pow(w_B_old(j,k),2)+1)),1);
          if (lr>log(R::runif(0,1))){
            ac_w2++;
            w_B_old(j,k)=w_new;
            PB_kernel_old.col(j)=PB_kernel_new.col(j);
            ll_z_old.col(j)=ll_z_new.col(j);
          }
        }
      }
    }
    
    //update zeta
    for (j=0;j<p;j++){
      count_zeta2++;
      zeta_new=R::rnorm(zeta_old(j),zeta_sd);
      for (i=0;i<n;i++){
        PB_kernel_new(i,j)=PB_kernel_old(i,j)+zeta_new-zeta_old(j);
        ll_z_new(i,j)=z(i,j)*PB_kernel_new(i,j)-logsumexp0(PB_kernel_new(i,j));
      }
      lr = sum(ll_z_new.col(j))-sum(ll_z_old.col(j))+R::dnorm4(zeta_new,log(1.0/99),zeta_sig,1)-R::dnorm4(zeta_old(j),log(1.0/99),zeta_sig,1);
      if (lr>log(R::runif(0,1))){
        ac_zeta2++;
        zeta_old(j)=zeta_new;
        PB_kernel_old.col(j)=PB_kernel_new.col(j);
        ll_z_old.col(j)=ll_z_new.col(j);
      }
    }
    
    
    //update beta
    d=accu(B_old);
    beta_old=R::rbeta(a_beta+d,b_beta+p*K_est-d);
    
    ll2(it)=accu(ll_z_old);
    if (it%1000==0){
      Rcpp::Rcout<<"postMCMC: "<<it<<"/"<<N2<<" completed"<<std::endl;
    }
    //write
    if (it>=burnin2&&it%thin2==0){
      B.slice(iter)=B_old;
      w_B.slice(iter)=w_B_old;
      zeta.col(iter)=zeta_old;
      beta(iter)=beta_old;
      iter++;
    }
  }
  ac_w2/=count_w2;
  ac_zeta2/=count_zeta2;

  arma::mat B_rate=mean(B,2);
  arma::umat B_est=(B_rate>.5);
  
  arma::mat w_B_est=mean(w_B,2);
  w_B_est = w_B_est%B_est;
  arma::vec zeta_est=mean(zeta,1);
  double beta_est=mean(beta);

  return Rcpp::List::create(
    Rcpp::List::create(
      Rcpp::Named("P_est")=P_est,
      Rcpp::Named("B_est")=B_est,
      Rcpp::Named("B_rate")=B_rate,
      Rcpp::Named("K_est")=K_est,
      Rcpp::Named("K_freq")=K_freq,
      Rcpp::Named("K")=K,
      Rcpp::Named("P")=P),
      Rcpp::List::create(
        Rcpp::Named("w_B_est")=w_B_est),
        Rcpp::List::create(
          Rcpp::Named("ac_w")=ac_w,
          Rcpp::Named("ac_zeta")=ac_zeta,
          Rcpp::Named("ll")=ll,
          Rcpp::Named("ll2")=ll2,
          Rcpp::Named("ll0")=ll0,
          Rcpp::Named("ac_w2")=ac_w2,
          Rcpp::Named("ac_zeta2")=ac_zeta2,
          Rcpp::Named("beta_est")=beta_est,
          Rcpp::Named("zeta_est")=zeta_est
        )
  );
}






